# base for protection layer views

from analyses.analysisbase import forms as app_forms
from analyses.analysisbase import tables as app_tables
from analyses.analysisbase.views.scenario import localnavigationlist
from analyses.analysisbase.views import helpers
from analyses.main import analyses_navigationlist

from django.core.urlresolvers import reverse
from django.db import models
from django.utils.translation import ugettext_lazy as _

from installations.views.main import installations_navigationlist

from utils import crud
from utils import permissions
from utils.forms import create_class_with_meta
from utils.versioning import versionedcrud
from utils.versioning.session import get_current_version
from utils.versioning.models import WorkVersion

class MixIn(object):
    "Common items for object views"
    include_attachments = True
    include_links = True
##    appconfig = app_models.AppConfig
##    model = app_models.DegradationProtectionLayer

class display(MixIn, versionedcrud.DisplayObjectView):
    form_class = None
    table_class = None
    def setup_from_args(self, *args, **kwargs):
        if not self.form_class:
            self.form_class = create_class_with_meta(app_forms.ProtectionLayerSheet, model=self.model)
        if not self.table_class:
            self.table_class = create_class_with_meta(app_tables.ProtectionLayerMeasuresTable, model=self.appconfig.get_model('protectionlayer_measure'))

    def add_to_extra_context(self, extra_context, object):
        if not object.SL:
            extra_context['navigationlist'] =  installations_navigationlist()
        extra_context['currentanalysistype'] = self.appconfig.analysistype_short
        versionedcrud.DisplayObjectView.add_to_extra_context(self,extra_context, object)
        breadcrumbs = [{'url': item.scenario.get_absolute_url('causetree'),
            'label': u'%s' % item.scenario} for item in object.get_treeelements() ]
        extra_context['breadcrumbs'] = breadcrumbs
        queryset = self.appconfig.get_model('protectionlayer_measure').objects.filter(protectionlayer=object).get_versionedobjects(version=object._version)
        table = self.table_class(queryset)
        table.title = _('Implementation: measures')
        if self.appconfig.get_models()['measure'].may_add(self.request.user):
            table.buttons = [{
                    'url': object.get_add_measure_url(),
                    'label': _('Add'),
                }]
        extra_context['table'] = table


class edit(MixIn, versionedcrud.EditObjectView):
    form_class = None

    def setup_from_args(self, *args, **kwargs):
        if not self.form_class:
            self.form_class = create_class_with_meta(app_forms.ProtectionLayerForm, model=self.model, me_model=self.appconfig.get_models()['measure'])

class delete(MixIn, versionedcrud.DeleteObjectView):
    form_class = None

    def setup_from_args(self, *args, **kwargs):
        if not self.form_class:
            self.form_class = create_class_with_meta(app_forms.ProtectionLayerSheet, model=self.model)

    def get_default_url(self, object):
        es = object.get_treeelements()
        if es:
            return es[0].scenario.get_absolute_url('causetree')
        return '/'

class add_to_scenario(MixIn, versionedcrud.AddObjectView):
    form_class = None

    def setup_from_args(self, object_id=None, *args, **kwargs):
        if not self.form_class:
            self.form_class = create_class_with_meta(app_forms.ProtectionLayerForm, model=self.model, me_model=self.appconfig.get_models()['measure'])
        # setting scenario here explicitly since it is not a parent of the protectionlayer
        # but of the treeelement
        # this gets set in the save step
        (self.version, self.specialversion) = get_current_version(self.request)
        self.scenario = versionedcrud.get_versionedobject_or_404(self.appconfig.get_models()['scenario'], self.version, v_id = object_id)
        treeparent_id = kwargs.pop('parent_id', None)
        if treeparent_id:
            try:
                self.treeparent = self.appconfig.get_models()['treeelement'].objects.get(pk=treeparent_id)
            except self.appconfig.get_models()['treeelement'].DoesNotExist:
                self.treeparent = None
        else:
            self.treeparent = None

    def check_permission(self, object):
        # override because we don't have a real parent defined
        if not self.appconfig.get_models()['treeelement'].may_add(self.request.user, parent=self.scenario):
            return permissions.redirect_to_not_allowed()

    def save_instance(self, object):
        object.save(self.request.user, parent=self.treeparent, treeelementkwargs={'scenario': self.scenario})
        return object

    def get_default_url(self, object):
        if self.scenario:
            return self.scenario.get_absolute_url('causetree')
        return '/'

class compare(MixIn, versionedcrud.CompareObjectView):
    form_class = None
    def setup_from_args(self, *args, **kwargs):
        if not self.form_class:
            self.form_class = create_class_with_meta(app_forms.ProtectionLayerSheet, model=self.model)
        super(compare, self).setup_from_args(*args, **kwargs)

def protectionlayer_extra_filter(queryset, data, appconfig):
    """
    Implement the special 'linked' filter which requires a very special query
    Used in filter and listobject
    """
    if data.get('linked'):
        if queryset:
            version = queryset[0]._version
            val = {u'True': True, u'False': False, True: True, False: False, 1: True, 0: False, u'1': True, u'0': False}[data['linked']]
            if val:
                crit = '> 0'
            else:
                crit = '= 0'
            t1 = '%s_%s' % (appconfig.namespace, appconfig.get_model('protectionlayer_measure').__name__.lower())
            t2 = '%s_%s' % (appconfig.namespace, appconfig.get_model('protectionlayer').__name__.lower())
            if version == WorkVersion:
                qs = queryset.extra(
                    where = ['''
                        (SELECT COUNT(*) from "%s" X
                        WHERE X."protectionlayer_id" = "%s"."v_id"
                        AND NOT (X."deleted" = %%s)
                        AND X."version_id" is null
                        ) %s''' %
                        (t1, t2, crit)], params=[True,]
                    ).get_versionedobjects(version)
            else:
                datev = models.DateTimeField().get_db_prep_value(version.date)
                qs = queryset.extra(
                    where = ['''
                        (SELECT COUNT(*) FROM "%s" X LEFT OUTER JOIN "versioning_version" Y ON X."version_id"=Y."id"
                        WHERE X."protectionlayer_id" = "%s"."v_id"
                        AND NOT (X."deleted" = %%s)
                        AND Y."date" =
                            (SELECT MAX(C."date")
                                FROM "%s" B LEFT OUTER JOIN "versioning_version" C on B."version_id"=C."id"
                                WHERE X."v_id" = B."v_id" and C."date" <= %%s)
                        ) %s''' %
                        (t1, t2, t1, crit)], params=[True, datev,]
                    ).get_versionedobjects(version)
            return qs, True
    return queryset, False

class ProtectionLayerFilter(crud.Filter):
    form_class = None
##    model = app_models.DegradationProtectionLayer
    filters = [
                {'filteritem': 'name', 'parameter': 'name__icontains', 'type': 'normal'},
            ]
    def extra_filter(self, queryset, data):
        return protectionlayer_extra_filter(queryset, data, self.appconfig)

    def get_form(self):
        if not self.form_class:
            return create_class_with_meta(app_forms.ProtectionLayerFilterForm, model=self.model)
        return self.form_class

class ProtectionLayerListObject(helpers.TableListObject):
    defaultsort = 'name'
    default_table_class = app_tables.ProtectionLayersTable

    def get_title(self):
        return u'%s %s' % (_('Protection layers'), self.appconfig.analysistype)

    def extra_filter(self, queryset, data):
        qs, filtered = protectionlayer_extra_filter(queryset, data, self.appconfig)
        return qs

class protectionlayers(helpers.TableListView):
    "List of protectionlayers"

##    filter = ProtectionLayerFilter()

    default_table_class = app_tables.ProtectionLayersTable
    defaultsort = 'name'
    paginate_by = 100
    redirect_on_unique = True

    def get_title(self):
        return u'%s %s' % (_('Protection layers'), self.appconfig.analysistype)

    def get_query_set(self, request):
        return self.model.objects.exclude(SL=True).get_versionedobjects(self.version)

    def add_to_extra_context(self, extra_context):
        super(protectionlayers, self).add_to_extra_context(extra_context)
        extra_context['navigationlist'] =  analyses_navigationlist(self.appconfig.namespace)
        extra_context['tablist'] =  self.get_navigationlist(self.appconfig.namespace)
        extra_context['currenttab'] = 'protectionlayers'
        extra_context['breadcrumbs'] = [{'label': _('Lists'), 'url': u''}]

    def get_navigationlist(self, namespace):
        return localnavigationlist(namespace)
