"""
Main module for analysis modules

Keeps a registry of analysis modules and provides interfaces to get info out of
the different modules.

"""
from django.core.urlresolvers import reverse
from django.utils.translation import ugettext_lazy as _
from itertools import groupby
from utils import forms as utils_forms


class AnalysesModules(object):
    """
    Registry for analysis modules.
    Every module needs to implement at least a scenario model and a measure
    model.
    To register the entire module, add the appconfig class to the registry.

    An appconfig must define:
        - namespace
        - analysistype
        - analysistype_short
        - get_models method

    """
    def __init__(self):
        self._registry = {}

    def register(self, appconfig, order=0):
        """
        Registers the given appconfig.
        """
        if appconfig in self.get_modules():
            # do nothing if already registered
            pass
        else:
            self._registry[appconfig.namespace] = {
                'appconfig': appconfig,
                'order': order}

    def __getitem__(self, name):
        """
        Make sure analysesmodules['somename'] returns the appconfig
        """
        return self._registry.get(name)['appconfig']

    def get_modules(self):
        return [item['appconfig'] for item in self.get_items_sorted()]

    def get_items_sorted(self):
        """
        Return the registry items sorted
        """
        res = self._registry.values()[:]
        res.sort(lambda x, y: cmp(x['order'], y['order']))
        return res

    def get_analysis_choices(self):
        """
        Get a list of tuples (name, analysistype) for use in choice fields
        """
        return [(item['appconfig'].namespace, item['appconfig'].analysistype)
            for item in self.get_items_sorted()]

    def get_analysis_add_choices(self, user):
        """
        Get a list of tuples (name, analysistype) for use in choice fields,
        limited to the types a user is allowed to add a scenario from
        """
        res = [('', '---------')]
        for item in self.get_items_sorted():
            if item['appconfig'].get_model('scenario').may_add(user):
                res.append((item['appconfig'].namespace,
                    item['appconfig'].analysistype))
        return res

    def show_analysis_add_choices(self, user):
        """
        Boolean whether the user has any rights to add a scenario
        """
        return len(self.get_analysis_add_choices(user)) > 1

    def get_report_formfields(self, user):
        """
        Return a structure of selection boxes which is a mix of the available
        selections for all analysis modules a user is allowed to display
        """
        res = []
        for item in self.get_items_sorted():
            scmodel = item['appconfig'].get_model('scenario')
            if scmodel.may_display(user):
                res = utils_forms.mix(res,
                    scmodel.get_listreport_formfields(user))
        return res


# create the singleton
analysesmodules = AnalysesModules()


def analyses_navigationlist(current=None, urlmethod = 'scenarios_url'):
    """
    Return a navigationlist of all registered analysis modules
    """
    items = analysesmodules.get_items_sorted()
    res = []
    for item in items:
        if item['appconfig'].namespace == current:
            res.append({'label': item['appconfig'].analysistype_short})
        else:
            url = reverse(getattr(item['appconfig'], urlmethod))
            res.append({'url': url, 'label': item['appconfig'].analysistype_short})
    return res


def get_scenarios_for_scopes(scopes, user, version=None):
    """
    Given a scope, return all the scenarios a user is allowed to see
    """
    scenarios = []
    if scopes:
        for module in analysesmodules.get_items_sorted():
            scmodel = module['appconfig'].get_model('scenario')
            if scmodel.may_display(user):
                sc = scmodel.objects.filter(scope__in=scopes
                    ).get_versionedobjects(version=version
                    ).select_related_versioned('scope')
                if sc:
                    scenarios.append(
                        {
                            'module': module,
                            'items': sc,
                        }
                    )
    return scenarios


def build_scenariotree(source, scope=None, analysistype=None):
    """
    Build a tree structure from a list of scenarios
    """
    scenarios = []
    for mod in source:
        module = mod['module']
        modscenarios = []
        for key, val in groupby(mod['items'], lambda x: x.type):
            d = {
                    'label': key and u'%s' % key or u'--',
                    'url': u'',
                    'type': u'scenariotype',
                }
            items = []
            for scenario in list(val):
                item = {
                        'url': scenario.get_absolute_url(),
                        'type': u'scenario',
                        'items': [],
                }
                if scope:
                    if scenario.scope == scope:
                        item['label'] = u'%s' % scenario
                    else:
                        scopetype = scenario.scope.auto and _('Section') or _('Scope')
                        item['label'] = u'%s (%s %s)' % (scenario, scopetype, scenario.scope)
                else:
                    item['label'] = u'%s' % scenario
                items.append(item)
            d['items'] = items
            modscenarios.append(d)
        scenarios.append(
            {
                'label': module['appconfig'].analysistype,
                'closed': analysistype and \
                    module['appconfig'].namespace != analysistype,
                'url': u'',
                'type': u'analysistype',
                'items': modscenarios,
            }
        )
    return scenarios


def get_scenarios_for_scope(scope, user, version=None):
    """
    Get the scenarios a user is allowed to see for a single scope
    """
    return get_scenarios_for_scopes([scope], user=user, version=version)


def collect_objects_for_scopes(scopes, version=None, filterkwarg={}):
    """
    Given a scope, use collect_objects on the scenarios of each type to get all
    objects
    """
    objects = []
    if scopes:
        for module in analysesmodules.get_items_sorted():
            scenarios = list(module['appconfig'].get_model('scenario'
                ).objects.filter(scope__in=scopes
                ).get_versionedobjects(version=version))
    # TODO: optimisation oportunity: collect for all scenarios at once
            for item in scenarios:
                objects.extend(item.collect_objects(filterkwarg))
    return objects


def collect_objects_for_scope(scope, version=None, filterkwarg={}):
    """
    Collect the objects for a single scope
    """
    return collect_objects_for_scopes([scope], version, filterkwarg)


def count_scenarios_for_scope(scope, version=None):
    """
    Count the scenarios for a scope
    """
    scenarios = 0
    for module in analysesmodules.get_items_sorted():
        sc = module['appconfig'].get_model('scenario'
            ).objects.filter(scope=scope).get_versionedobjects(version=version)
        scenarios += len(sc)
    return scenarios


def get_measures(filters, user, version=None):
    measures = []
    for module in analysesmodules.get_items_sorted():
        memodel = module['appconfig'].get_model('measure')
        if memodel.may_display(user):
            me = list(memodel.objects.filter(**filters
                ).get_versionedobjects(version=version))
            if me:
                measures.extend(me)
    return measures


def get_urls():
    """
    Create the url configuration for all registered modules
    """
    from django.conf.urls.defaults import patterns, url, include
    urlpatterns = patterns('',)
    for module in analysesmodules.get_items_sorted():
        try:
            urlpatterns += patterns('',
                url(r'^', include(module['appconfig'].get_urls())))
        except:
            pass
    return urlpatterns
