"""
replacements for ModelForm, ModelFormset and InlineFormset
"""

from django import forms
from django.db import models

from fields import DateFormField
from widgets import DateWidget

class ModelFormMetaclass(forms.models.ModelFormMetaclass):
    """
    Special form metaclass
    Allows specifying the following items directly in the Meta class of the form:
        - widgets
        - attrs
        - labels
        - help_texts
        - fieldargs
    This avoids having to do these changes in the form's init
    """
    def __new__(cls, name, bases, attrs):
        meta = attrs.get('Meta', None)
        _widgets = getattr(meta, 'widgets', {})
        _attrs = getattr(meta, 'attrs', {})
        _labels = getattr(meta, 'labels', {})
        _help_texts = getattr(meta, 'help_texts', {})
        _fieldargs = getattr(meta, 'fieldargs', {})

        def formfield_callback(f, **kwargs):
            # use special Date stuff
            if isinstance(f, models.DateField):
                kwargs["widget"] = DateWidget
                return f.formfield(form_class=DateFormField, **kwargs)
            # allow overriding of widgets in Meta
            if f.name in _widgets:
                kwargs["widget"] = _widgets[f.name]
            # allow overriding of attrs in Meta
            if f.name in _attrs:
                widget = kwargs.pop("widget", f.formfield().widget)
                if widget :
    				widget.attrs.update(_attrs[f.name])
    				kwargs["widget"] = widget
            # allow overriding of labels in Meta
            if f.name in _labels:
                kwargs["label"] = _labels[f.name]
            # allow overriding of help_texts in Meta
            if f.name in _help_texts:
                kwargs["help_text"] = _help_texts[f.name]
            # allow adding extra field arguments
            if f.name in _fieldargs:
                kwargs.update(_fieldargs[f.name])
            return f.formfield(**kwargs)
        attrs['formfield_callback'] = formfield_callback

        return super(ModelFormMetaclass, cls).__new__(cls, name, bases,
                attrs)

class ModelForm(forms.ModelForm):
    __metaclass__ = ModelFormMetaclass

class ModelFormSetMetaclass(type):
    """
    Special formset metaclass
    Used to avoid using the factory function for modelformsets
    Accepts the following attributes directly in the formset class:
        - can_order (default = False)
        - can_delete (default = False)
        - max_num (default = 0)
        - extra (default = 1)
    Required attributes:
        - form
        - model
    """
    def __new__(cls, name, bases, attrs):
        try:
            parents = [b for b in bases if issubclass(b, ModelFormSet)]
        except NameError:
            # We are defining ModelFormSet itself.
            parents = None
        new_class = super(ModelFormSetMetaclass, cls).__new__(cls, name, bases,
                            attrs)
        if not parents:
            return new_class

        if 'can_order' not in attrs:
            new_class.can_order = False
        if 'can_delete' not in attrs:
            new_class.can_delete = False
        if 'max_num' not in attrs:
            new_class.max_num = None
        if 'extra' not in attrs:
            new_class.extra = 1

        if 'form' not in attrs:
            raise Exception(u'Error in definition of ModelFormSet subclass %s. You need to define a form.' % name)

        if 'model' not in attrs:
            raise Exception(u'Error in definition of ModelFormSet subclass %s. You need to define a model.' % name)
        model = attrs.get('model')

        new_class.model = model

        return new_class

class ModelFormSet(forms.models.BaseModelFormSet):
    """
    ModelFormSet subclasses need these attributes: form and model
    """
    __metaclass__ = ModelFormSetMetaclass

class InlineFormSetMetaclass(type):
    """
    Special inline formset metaclass
    Used to avoid using the factory function for inline modelformsets
    Accepts the following attributes directly in the formset class:
        - can_order (default = False)
        - can_delete (default = False)
        - max_num (default = 0)
        - extra (default = 1)
    Required attributes:
        - form
        - model
        - parent (name of foreign key field to parent)
    """
    def __new__(cls, name, bases, attrs):
        try:
            parents = [b for b in bases if issubclass(b, InlineFormSet)]
        except NameError:
            # We are defining InlineFormSet itself.
            parents = None
        new_class = super(InlineFormSetMetaclass, cls).__new__(cls, name, bases,
                            attrs)
        if not parents:
            return new_class

        if 'can_order' not in attrs:
            new_class.can_order = False
        if 'can_delete' not in attrs:
            new_class.can_delete = False
        if 'max_num' not in attrs:
            new_class.max_num = None
        if 'extra' not in attrs:
            new_class.extra = 3
        try:
            form = attrs.get('form')
        except:
            raise Exception(u'Error in definition of InlineFormset subclass %s. You need to define a form' % name)

        try:
            model = attrs.get('model')
        except:
            raise Exception(u'Error in definition of InlineFormset subclass %s. You need to define a model' % name)

        try:
            parent = attrs.pop('parent')
        except:
            raise Exception(u'Error in definition of InlineFormset subclass %s. You need to define a parent attribute' % name)

        try:
            fk = (i for i in model._meta.fields if i.name == parent).next()
        except:
            raise Exception(u'Error in definition of InlineFormset subclass %s. The parent attribute of the inlineformset is not a valid field' % name)

        if not isinstance(fk, models.ForeignKey):
            raise Exception(u'Error in definition of InlineFormset subclass %s. The parent attribute is not a ForeignKey' % name)

        new_class.fk = fk

        return new_class

class InlineFormSet(forms.models.BaseInlineFormSet):
    __metaclass__ = InlineFormSetMetaclass



def create_class_with_meta(parent_class, **meta_attrs):
    """
    Dynamic class creation, adding a model to the Meta information.
    Can be used for form, sheet or table definitions
    """
    parents = (object,)
    if hasattr(parent_class, 'Meta'):
        parents = (parent_class.Meta, object)
    Meta = type('Meta', parents, meta_attrs)
    class_name =parent_class.__name__ + 'Child'
    class_attrs = {
        'Meta': Meta,
    }
    return type(class_name, (parent_class,), class_attrs)

def flatten(struct, level):
    res = []
    for item in struct:
        res.append({'name': item['name'], 'field': item['field'], 'level': level})
        children = item.get('children')
        if children:
            res.extend(flatten(children, level+1))
    return res

def treeify(items):
    "Iterator, adding new_level and closed_level attributes"
    from mptt.utils import previous_current_next
    items = flatten(items, 0)
    for previous, current, next in previous_current_next(items):
        current_level = current['level']
        if previous:
            current['field'].new_level = (previous['level'] < current_level)
            current['new_level'] = (previous['level'] < current_level)
        else:
            current['field'].new_level = True
            current['new_level'] = True
        if next:
            current['field'].closed_levels = range(current_level, next['level'], -1)
            current['closed_levels'] = range(current_level, next['level'], -1)
        else:
            current['field'].closed_levels = range(current_level, -1, -1)
            current['closed_levels'] = range(current_level, -1, -1)
        yield current

def mix(struct1, struct2):
    res = []
    names2 ={}
    for item in struct2:
        names2[item['name']] = item
    for item in struct1:
        found = False
        for item2 in struct2:
            if item['name'] == item2['name']:
                found = True
                if item.get('children'):
                    if item2.get('children'):
                        ch = mix(item.get('children'), item2.get('children'))
                    else:
                        ch = mix([], item.get('children'))
                else:
                    if item2.get('children'):
                        ch = mix([], item2.get('children'))
                    else:
                        ch = []
        if not found:
            ch = mix([], item.get('children'))
        res.append({'name': item['name'], 'field': item['field'], 'children': ch})
    for item2 in struct2:
        found = False
        for item in struct1:
            if item['name'] == item2['name']:
                found = True
        if not found:
            res.append({'name': item2['name'], 'field': item2['field'], 'children': mix([], item2.get('children'))})
    return res

class ReportForm(forms.Form):
    def __init__(self, user, *args, **kwargs):
        super(ReportForm, self).__init__(*args, **kwargs)
        fields = treeify(self.Meta.model.get_report_formfields(user))
        for k in fields:
            self.fields[k['name']] = k['field']

class ListReportForm(forms.Form):
    def __init__(self, user, *args, **kwargs):
        super(ListReportForm, self).__init__(*args, **kwargs)
        fields = treeify(self.Meta.model.get_listreport_formfields(user))
        for k in fields:
            self.fields[k['name']] = k['field']
